from __future__ import annotations

import asyncio
from collections.abc import Hashable

import discord
from discord.ext import commands

from .bot import StarBot, StarState
from .db import AbstractConnection


class Locks:
    def __init__(self) -> None:
        self.locks: dict[Hashable, asyncio.Lock] = {}

    def lock_for(self, key: Hashable) -> asyncio.Lock:
        if key in self.locks:
            return self.locks[key]
        else:
            return self.locks.setdefault(key, asyncio.Lock())


locks = Locks()
lock_for = locks.lock_for


class StarCtx:
    def __init__(self, ctx: commands.Context) -> None:
        self.ctx = ctx
        assert ctx.guild
        self.guild: discord.Guild = ctx.guild
        self.bot: StarBot = ctx.bot
        self.state: StarState = self.bot.starstate
        self.connection: AbstractConnection = self.state.connection

    async def assign(self, count: int) -> None:
        await self.connection.set(("assign", self.guild.id), {"channel": self.ctx.channel.id, "count": count})

    async def unassign(self) -> None:
        await self.connection.set(("assign", self.guild.id), None)


class ReactionCtx:
    def __init__(self, bot: StarBot, event: discord.RawReactionActionEvent):
        self.bot = bot
        self.guild_id = event.guild_id
        self.channel_id = event.channel_id
        self.message_id = event.message_id
        self.name = event.emoji.name
        self.state: StarState = self.bot.starstate
        self.connection: AbstractConnection = self.state.connection

    async def get_channel(self, id_: int) -> discord.abc.Messageable:
        channel = self.bot.get_channel(id_)
        if channel is None:
            channel = await self.bot.fetch_channel(id_)
        if isinstance(channel, discord.CategoryChannel):
            raise TypeError
        if isinstance(channel, discord.ForumChannel):
            raise TypeError
        if isinstance(channel, discord.abc.PrivateChannel):
            raise TypeError
        return channel

    async def on(self) -> None:
        if self.name != "⭐":
            return
        assignment: dict[str, int] | None = self.connection.get(("assign", self.guild_id), None)
        if assignment is None:
            return
        assigned_to, count = assignment["channel"], assignment["count"]
        if self.channel_id == assigned_to:
            return
        async with lock_for(self.message_id):
            assigned_channel, event_channel = await asyncio.gather(
                self.get_channel(assigned_to), self.get_channel(self.channel_id)
            )
            message = await event_channel.fetch_message(self.message_id)
            reaction = next((reaction for reaction in message.reactions if reaction.emoji == "⭐"), None)
            if reaction is None:
                return
            if reaction.me:
                return
            if reaction.count >= count:
                guild = message.guild
                if guild is None:
                    return
                member = guild.get_member(message.author.id)
                if member is None:
                    member = await guild.fetch_member(message.author.id)
                embed = discord.Embed(description=message.content or None)
                avatar = member.avatar
                embed.set_author(name=member.display_name, url=message.jump_url, icon_url=avatar and avatar.url)
                image = next(
                    (
                        attachment.url
                        for attachment in message.attachments
                        if (attachment.content_type or "").startswith("image/")
                    ),
                    None,
                )
                if image is not None:
                    embed.set_image(url=image)
                await asyncio.gather(
                    message.add_reaction("⭐"),
                    assigned_channel.send(embed=embed),
                )


class Stars(commands.Cog):
    def __init__(self, bot: StarBot) -> None:
        super().__init__()
        self.__bot = bot

    @commands.hybrid_command()
    async def ping(self, ctx: commands.Context):
        print("ping pong")
        await ctx.reply("pong", mention_author=False)

    @commands.hybrid_command()
    @commands.is_owner()
    async def reload(self, ctx: commands.Context):
        print("reload")
        bot: commands.Bot = ctx.bot
        try:
            await bot.reload_extension("starbot.stars")
        except commands.ExtensionNotLoaded:
            await ctx.reply("not loaded")
        print("reloaded")
        await ctx.reply("reloaded")

    @commands.hybrid_command()
    @commands.is_owner()
    async def sync(self, ctx: commands.Context):
        await ctx.bot.tree.sync()
        print("synced")
        await ctx.reply("synced")

    @commands.hybrid_command()
    @commands.has_permissions(administrator=True)
    async def assign(self, ctx: commands.Context, count: int):
        await StarCtx(ctx).assign(count)
        await ctx.reply("assigned")

    @commands.hybrid_command()
    @commands.has_permissions(administrator=True)
    async def unassign(self, ctx: commands.Context):
        await StarCtx(ctx).unassign()
        await ctx.reply("unassigned")

    @commands.Cog.listener()
    async def on_raw_reaction_add(self, event: discord.RawReactionActionEvent):
        await ReactionCtx(self.__bot, event).on()


async def setup(bot: StarBot):
    global cog
    cog = Stars(bot)
    await bot.add_cog(cog)


async def teardown(bot: StarBot):
    global cog
    await bot.remove_cog(cog.qualified_name)
    del cog
    print("torn down")
